""" Blends forecasts together

1. creates weights for blending
2. get forecast values for each model
3. blends them together

"""

import json
from datetime import datetime
from typing import List, Optional

import pandas as pd
import structlog
from sqlalchemy.orm.session import Session

from nowcasting_datamodel.models.forecast import ForecastValue, ForecastValueSevenDaysSQL
from nowcasting_datamodel.read.blend.utils import (
    blend_forecasts_together,
    check_forecast_created_utc,
    convert_df_to_list_forecast_values,
    convert_list_forecast_values_to_df,
)
from nowcasting_datamodel.read.blend.weights import (
    make_weights_df,
)
from nowcasting_datamodel.read.read import get_forecast_values, get_forecast_values_latest

logger = structlog.stdlib.get_logger()


def get_blend_forecast_values_latest(
    session: Session,
    gsp_id: int,
    start_datetime: Optional[datetime] = None,
    model_names: Optional[List[str]] = None,
    weights: Optional[List[float]] = None,
    forecast_horizon_minutes: Optional[int] = None,
    properties_model: Optional[str] = None,
) -> List[ForecastValue]:
    """
    Get forecast values

    :param session: database session
    :param gsp_id: gsp id, to filter query on
    :param start_datetime: optional to filterer target_time by start_datetime
        If None is given then all are returned.
    :param model_names: list of model names to use for blending
    :param weights: list of weights to use for blending, see structure in make_weights_df
    :param properties_model: the model to use for the properties

    return: List of forecasts values blended from different models
    """

    logger.info(f"Getting blend forecast for gsp_id {gsp_id} and start_datetime {start_datetime}")

    if model_names is None:
        model_names = ["cnn", "National_xg"]
    if len(model_names) > 1:
        weights_df = make_weights_df(
            model_names, weights, start_datetime, forecast_horizon_minutes=forecast_horizon_minutes
        )
    else:
        weights_df = None

    if properties_model is not None:
        assert (
            properties_model in model_names
        ), f"properties_model must be in model_names {model_names}"

    # get forecast for the different models
    forecast_values_all_model = []
    for model_name in model_names:
        if forecast_horizon_minutes is None:
            forecast_values_one_model = get_forecast_values_latest(
                session=session, gsp_id=gsp_id, start_datetime=start_datetime, model_name=model_name
            )
        else:
            forecast_values_one_model = get_forecast_values(
                session=session,
                gsp_ids=[gsp_id],
                start_datetime=start_datetime,
                only_return_latest=True,
                forecast_horizon_minutes=forecast_horizon_minutes,
                model=ForecastValueSevenDaysSQL,
                model_name=model_name,
            )

        if len(forecast_values_one_model) == 0:
            logger.debug(
                f"No forecast values for {model_name} for gsp_id {gsp_id} "
                f"and start_datetime {start_datetime}"
            )
        else:
            logger.debug(
                f"Found {len(forecast_values_one_model)} values for {model_name} "
                f"for gsp_id {gsp_id} and start_datetime {start_datetime}"
            )
            forecast_values_all_model.append([model_name, forecast_values_one_model])

    # check the created_utc is valid for each forecast
    forecast_values_all_model_valid = check_forecast_created_utc(forecast_values_all_model)

    # make into dataframe
    forecast_values_all_model = convert_list_forecast_values_to_df(forecast_values_all_model_valid)

    # blend together
    forecast_values_blended = blend_forecasts_together(forecast_values_all_model, weights_df)

    # add properties
    forecast_values_df = add_properties_to_forecast_values(
        blended_df=forecast_values_blended,
        properties_model=properties_model,
        all_model_df=forecast_values_all_model,
    )

    # convert back to list of forecast values
    forecast_values = convert_df_to_list_forecast_values(forecast_values_df)

    return forecast_values


def add_properties_to_forecast_values(
    blended_df: pd.DataFrame,
    all_model_df: pd.DataFrame,
    properties_model: Optional[str] = None,
):
    """
    Add properties to blended forecast values, we just take it from one model.

    We normalize all properties by the "expected_power_generation_megawatts" value,
    and renormalize by the blended "expected_power_generation_megawatts" value.
    This makes sure that plevels 10 and 90 surround the blended value.

    :param blended_df: dataframe of blended forecast values
    :param properties_model: which model to use for properties
    :param all_model_df: dataframe of all forecast values for all models
    :return:
    """

    logger.debug(
        f"Adding properties to blended forecast values for properties_model {properties_model}"
    )

    if properties_model is None:
        blended_df["properties"] = None
        return blended_df

    # get properties

    properties_df = all_model_df[all_model_df["model_name"] == properties_model]
    properties_df.reset_index(inplace=True, drop=True)

    # adjust "properties" to be relative to the expected_power_generation_megawatts
    # this is a bit tricky becasue the "properties" column is a list of dictionaries
    # below we add "expected_power_generation_megawatts" value back to this.
    # We do this so that plevels are relative to the blended values.
    properties_only_df = pd.json_normalize(properties_df["properties"])
    for c in properties_only_df.columns:
        properties_only_df[c] -= properties_df["expected_power_generation_megawatts"]
    properties_df["properties"] = properties_only_df.apply(
        lambda x: json.loads(x.to_json()), axis=1
    )

    # reduce columns
    properties_df = properties_df[["target_time", "properties"]]

    # add properties to blended forecast values
    blended_df = blended_df.merge(properties_df, on=["target_time"], how="left")

    # add "expected_power_generation_megawatts" to the properties
    properties_only_df = pd.json_normalize(blended_df["properties"])
    for c in properties_only_df.columns:
        properties_only_df[c] += blended_df["expected_power_generation_megawatts"]
    blended_df["properties"] = properties_only_df.apply(lambda x: json.loads(x.to_json()), axis=1)

    assert "properties" in blended_df.columns

    return blended_df



""" Utils functions for test """
from datetime import datetime, timezone

# Used constants
from nowcasting_datamodel.utils import convert_to_camelcase, datetime_must_have_timezone


def test_datetime_must_have_timezone():
    """Test function datetime_must_have_timezone"""

    time_now = datetime.now(timezone.utc)

    # check functions works
    new_time_now = datetime_must_have_timezone(None, time_now)
    assert new_time_now == time_now

    # check functions adds timezone
    time_now = datetime.now()
    time_now = datetime_must_have_timezone(None, time_now)
    assert new_time_now != time_now
    assert new_time_now.tzinfo == timezone.utc


def test_convert_to_camelcase():
    """Test convert to camelcase works"""
    assert convert_to_camelcase("foo_bar") == "fooBar"
    assert convert_to_camelcase("foo_bar_baz") == "fooBarBaz"



""" Make national forecasts """
import logging
from typing import List

import pandas as pd
from sqlalchemy.orm.session import Session

from nowcasting_datamodel import N_GSP
from nowcasting_datamodel.models import (
    Forecast,
    ForecastSQL,
    ForecastValue,
    MLModel,
    national_gb_label,
)
from nowcasting_datamodel.read.read import get_latest_input_data_last_updated, get_location

logger = logging.getLogger(__name__)


def make_national_forecast(
    forecasts: List[ForecastSQL], session: Session, n_gsps: int = N_GSP
) -> ForecastSQL:
    """This takes a list of forecast and adds up all the forecast values

    Note that a different method to do this, would be to do this in the database
    """
    assert (
        len(forecasts) == n_gsps
    ), f"The number of forecast was only {len(forecasts)}, it should be {n_gsps}"

    # check the gsp ids are unique
    gsps = [int(f.location.gsp_id) for f in forecasts]
    seen = set()
    dupes = [x for x in gsps if x in seen or seen.add(x)]
    assert len(seen) == n_gsps, f"Found non unique GSP ids {dupes}"

    location = get_location(gsp_id=0, session=session, label=national_gb_label)
    input_data_last_updated = get_latest_input_data_last_updated(session=session)

    # make pandas dataframe of all the forecast values with a gsp id
    forecast_values_flat = []
    for forecast in forecasts:
        gsp_id = forecast.location.gsp_id

        one_gsp = pd.DataFrame(
            [ForecastValue.from_orm(value).dict() for value in forecast.forecast_values]
        )
        adjusts_mw = [f.adjust_mw for f in forecast.forecast_values]
        one_gsp["gps_id"] = gsp_id
        one_gsp["_adjust_mw"] = adjusts_mw
        # this is a data frame with the follow columns
        # - gsp_id
        # - target_time
        # - expected_power_generation_megawatts

        forecast_values_flat.append(one_gsp)

    forecast_values_df = pd.concat(forecast_values_flat)
    forecast_values_df[
        "count"
    ] = 1  # this will be used later to check that there are 338 forecasts for each target_time

    # group by target time
    forecast_values_df = forecast_values_df.groupby(["target_time"]).sum()
    forecast_values_df.reset_index(inplace=True)
    if (forecast_values_df["count"] != n_gsps).all():
        m = f'The should should be {n_gsps}, but instead it is {forecast_values_df["count"]}'
        logger.debug(m)
        raise Exception(m)

    # change back to ForecastValue
    forecast_values = []
    for _, row in forecast_values_df.iterrows():
        f = ForecastValue(
            target_time=row["target_time"],
            expected_power_generation_megawatts=row["expected_power_generation_megawatts"],
        )
        f._adjust_mw = row["_adjust_mw"]
        forecast_values.append(f)

    # change to ForecastValueSQL
    forecast_values = [forecast_value.to_orm() for forecast_value in forecast_values]

    # change to sql
    model = forecasts[0].model
    if isinstance(model, MLModel):
        model = model.to_orm()

    national_forecast = ForecastSQL(
        forecast_values=forecast_values,
        location=location,
        model=model,
        input_data_last_updated=input_data_last_updated,
        forecast_creation_time=forecasts[0].forecast_creation_time,
        historic=False,
    )

    # validate
    _ = Forecast.from_orm(national_forecast)

    return national_forecast



"""Utils for blending forecasts together"""
from datetime import datetime, timedelta, timezone
from typing import List, Union

import numpy as np
import pandas as pd
import structlog

from nowcasting_datamodel.models import ForecastValue

logger = structlog.stdlib.get_logger()


def check_forecast_created_utc(forecast_values_all_model) -> List[Union[str, List[ForecastValue]]]:
    """
    Check if forecasts are valid.

    We only consider forecast less than 2 hours old.
    If all forecast are older than 2 hours, we used both.

    :param forecast_values_all_model: list of forecast
    :return: list of valid forecasts
    """
    one_forecast_created_within_timedelta = False

    # if all forecasts are later than 2 hours, then we use the blend,
    # otherwise we only use forecast less than 2 hours
    if one_forecast_created_within_timedelta:
        # remove all forecasts that are older than 2 hours
        forecast_values_all_model_valid = []
        for model_name, forecast_values_one_model in forecast_values_all_model:
            one_forecast_created_within_timedelta = forecast_values_one_model[
                0
            ].created_utc < datetime.now(timezone.utc) - timedelta(hours=2)

            if one_forecast_created_within_timedelta:
                logger.debug(f"Will be using forecast {model_name} " f"as it is newer than 2 hours")
                forecast_values_all_model_valid.append([model_name, forecast_values_one_model])
            else:
                logger.debug(f"forecast {model_name} is older than 2 hours, so not using it")
    else:
        # use all forecast:
        logger.debug("using all forecasts as all are older than 2 hours")
        forecast_values_all_model_valid = forecast_values_all_model
    return forecast_values_all_model_valid


def convert_list_forecast_values_to_df(forecast_values_all_model_valid):
    """
    Convert list of forecast values to a pandas dataframe

    :param forecast_values_all_model_valid:
    :return: merged into pandas dataframe with columns
     - "expected_power_generation_megawatts",
     - "adjust_mw",
     - "created_utc",
     - "model_name",
    """
    logger.debug(f"Getting values from {len(forecast_values_all_model_valid)} models")
    forecast_values_all_model_df = []
    for model_name, forecast_values_one_model in forecast_values_all_model_valid:
        logger.debug(f"Making dataframe for {model_name}")

        values_list = [
            [
                value.target_time,
                value.expected_power_generation_megawatts,
                value.adjust_mw,
                value.created_utc,
                value.properties,
                model_name,
            ]
            for value in forecast_values_one_model
        ]
        value_df = pd.DataFrame(
            values_list,
            columns=[
                "target_time",
                "expected_power_generation_megawatts",
                "adjust_mw",
                "created_utc",
                "properties",
                "model_name",
            ],
        )
        value_df.reset_index(inplace=True, drop=False)
        forecast_values_all_model_df.append(value_df)
    # join into one dataframe
    forecast_values_all_model = pd.concat(forecast_values_all_model_df, axis=0)
    forecast_values_all_model.reset_index(inplace=True)

    return forecast_values_all_model


def convert_df_to_list_forecast_values(forecast_values_blended: pd.DataFrame):
    """
    Convert the blended dataframe to a list of ForecastValue objects

    :param forecast_values_blended: Needs to have the columns 'target_time'
        and 'expected_power_generation_megawatts
    :return:
    """
    forecast_values = []
    logger.debug(forecast_values_blended)
    for i, row in forecast_values_blended.iterrows():
        # make sure we don't have negative values
        expected_power_generation_megawatts = row.expected_power_generation_megawatts
        if np.isnan(expected_power_generation_megawatts):
            continue
        if expected_power_generation_megawatts < 0:
            expected_power_generation_megawatts = 0

        forecast_value = ForecastValue(
            target_time=row.target_time,
            expected_power_generation_megawatts=expected_power_generation_megawatts,
        )
        forecast_value._adjust_mw = row.adjust_mw
        forecast_value._properties = row.properties
        forecast_values.append(forecast_value)
    return forecast_values


def blend_forecasts_together(forecast_values_all_model, weights_df):
    """
    Blend the forecasts together using the weights_df

    :param forecast_values_all_model: Dataframe containing the columns 'target_time',
        'expected_power_generation_megawatts', 'adjust_mw', 'model_name'
    :param weights_df: Dataframe of weights with columns 'model_name' and 'weight'
    :return: Dataframe with the columns
        'target_time',
        'expected_power_generation_megawatts',
        'adjust_mw'
    """

    # drop of the "properties" column
    if "properties" in forecast_values_all_model.columns:
        forecast_values_all_model = forecast_values_all_model.drop(columns=["properties"]).copy()

    # blend together
    # lets deal with unique target times first
    logger.debug(forecast_values_all_model["target_time"])
    # get all unique target times
    all_target_times = forecast_values_all_model["target_time"].unique()
    logger.debug(f"Found in total {len(all_target_times)} target times")

    # get the duplicated target times
    duplicated_target_times = forecast_values_all_model[
        forecast_values_all_model["target_time"].duplicated()
    ]["target_time"].tolist()
    logger.debug(f"Found {len(duplicated_target_times)} duplicated target times")
    unique_target_times = [x for x in all_target_times if x not in duplicated_target_times]
    logger.debug(f"Found in {len(unique_target_times)} unique target times")

    # get the index of the duplicated and unique target times
    duplicated_target_times_idx = forecast_values_all_model[
        forecast_values_all_model["target_time"].isin(duplicated_target_times)
    ].index
    unique_target_times_idx = forecast_values_all_model[
        forecast_values_all_model["target_time"].isin(unique_target_times)
    ].index

    # get the unique forecast values
    forecast_values_blended = forecast_values_all_model.loc[unique_target_times_idx]

    # now lets deal with the weights
    duplicated = forecast_values_all_model.loc[duplicated_target_times_idx]
    duplicated = duplicated.drop_duplicates()

    # only do this if there are duplicated
    if len(duplicated) > 0:
        logger.debug(f"Now blending the duplicated target times using {weights_df}")

        pd.DataFrame()
        # unstack the weights
        weights_one_df = weights_df.stack()
        weights_one_df.name = "weight"
        weights_one_df = weights_one_df.reset_index()
        weights_one_df = weights_one_df.rename(columns={"level_1": "model_name"})
        weights_one_df = weights_one_df.rename(columns={"level_0": "target_time"})

        # join the weights to the duplicated
        duplicated = pd.merge(
            duplicated,
            weights_one_df,
            how="left",
            left_on=["model_name", "target_time"],
            right_on=["model_name", "target_time"],
        )

        # multiply the expected power generation by the weight
        for col in ["expected_power_generation_megawatts", "adjust_mw"]:
            duplicated[col] = duplicated[col] * duplicated["weight"]
        duplicated.drop(columns=["created_utc"], inplace=True)

        # sum the weights
        duplicated = duplicated.groupby(["target_time"]).sum()

        # make sure target_time is a columns
        duplicated["target_time"] = duplicated.index
        duplicated.reset_index(inplace=True, drop=True)

        # divide by the sum of the weights, # TODO should we be worried about dividing by zero?
        for col in ["expected_power_generation_megawatts", "adjust_mw"]:
            duplicated[col] /= duplicated["weight"]

        logger.debug(duplicated)
    # join unique and duplicates together
    forecast_values_blended = pd.concat([forecast_values_blended, duplicated], axis=0)
    # sort by target time
    forecast_values_blended.sort_values(by="target_time", inplace=True)
    return forecast_values_blended
